from __future__ import annotations

from datetime import datetime
from typing import List

from pydantic import BaseModel, Field
from typing_extensions import Annotated


class Version(BaseModel):
    tag: int
    created_at: datetime = Field(validation_alias="createdAt")
    created_by: str = Field(validation_alias="createdBy")
    base_model: str = Field(validation_alias="baseModel")
    archived: bool


class Repo(BaseModel):
    uuid: str
    name: str
    description: str | None = Field(default=None)
    all_versions: Annotated[List[Version], Field(default_factory=list, validation_alias="versions", repr=False)]

    @property
    def versions(self, show_archived: bool = False) -> List[Version]:
        return [v for v in self.all_versions if show_archived or not v.archived]



from typing import Optional

import typer

from predibase.cli_commands.utils import get_client

app = typer.Typer(no_args_is_help=True)


@app.command(help="Fine-tune a Large Language Model (LLM)")
def llm(
    base_model_name: str = typer.Option(
        None,
        "--base-model",
        "-b",
        prompt="Name of the base model to fine-tune",
        prompt_required=True,
        help="Name of the base model to fine-tune",
    ),
    repo_name: Optional[str] = typer.Option(
        None,
        "--repo-name",
        "-r",
        prompt="Name of the repo in which to save the new model",
        prompt_required=True,
        help="Name of the repo in which to save the new model",
    ),
    prompt_template: str = typer.Option(
        None,
        "--prompt-template",
        prompt="Template input",
        prompt_required=True,
        help="Prompt template input text",
    ),
    target: str = typer.Option(
        None,
        "--target",
        prompt="Fine-tuning target column",
        prompt_required=True,
        help="Fine-tuning target column",
    ),
    dataset_name: Optional[str] = typer.Option(
        None,
        "--dataset",
        "-d",
        prompt="Dataset to fine-tune on",
        prompt_required=True,
        help="Dataset to fine-tune on",
    ),
    template_name: str = typer.Option(
        None,
        "--template",
        prompt="Fine-tuning template to use",
        prompt_required=False,
        help="Fine-tuning template to use (optional)",
    ),
    epochs: Optional[int] = typer.Option(
        None,
        "--epochs",
        prompt="Number of epochs to train for",
        prompt_required=False,
        help="Number of epochs to train for",
    ),
    train_steps: Optional[int] = typer.Option(
        None,
        "--steps",
        prompt="Number of steps to train for",
        prompt_required=False,
        help="Number of steps to train for",
    ),
    learning_rate: Optional[float] = typer.Option(
        None,
        "--lr",
        prompt="Learning rate to use",
        prompt_required=False,
        help="Learning rate to use",
    ),
    wait: Optional[bool] = typer.Option(
        None,
        "--wait",
        prompt="Whether to wait until training finishes",
        prompt_required=False,
        help="If set, the deploy command will not return until the training process finishes",
    ),
):
    client = get_client()
    tmpls = client.LLM(base_model_name).get_finetune_templates()
    tmpl = tmpls.default
    if template_name:
        try:
            tmpl = tmpls[template_name]
        except KeyError:
            raise ValueError(f"provided template {template_name} does not exist") from None

    job = tmpl.run(
        prompt_template=prompt_template,
        target=target,
        dataset=dataset_name,
        repo=repo_name,
        epochs=epochs,
        train_steps=train_steps,
        learning_rate=learning_rate,
    )

    if wait:
        job.get()


@app.command(help="Get possible templates for fine-tuning a Large Language Model (LLM)")
def templates(
    base_model_name: str = typer.Option(
        None,
        "--base-model",
        "-b",
        prompt="Name of the base model to fine-tune",
        prompt_required=True,
        help="Name of the base model to fine-tune",
    ),
):
    client = get_client()
    tmpls = client.LLM(base_model_name).get_finetune_templates()
    tmpls.compare()


if __name__ == "__main__":
    app()



from dataclasses import dataclass, field
from typing import Any, Dict

import yaml
from dataclasses_json import dataclass_json, LetterCase


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ConfigSuggestion:
    """Describes a ludwig configuration suggestion."""

    # Why this configuration is useful or interesting to try.
    description: str = ""

    # The raw ludwig config, without backend or execution-specific parameters.
    config: Dict[str, Any] = field(default_factory=dict)

    # The model category this config is designed for, i.e. TABULAR, IMAGE, TEXT, TEXT_TABULAR, MULTI_MODAL, etc, largely
    # describes the nature of the model's input features.
    model_category: str = "UNKNOWN"

    # Whether the config includes a hyperopt section, and should run a hyperparameter search.
    contains_hyperopt: bool = False

    # Whether the config uses a pretrained model.
    uses_pretrained_model: bool = False

    def __repr__(self) -> str:
        return yaml.dump(self.to_dict())



from typing import Tuple

import os
import openai


def parse_connection_and_dataset_name(s: str) -> Tuple[str, str]:
    segments = s.split("/")
    if len(segments) == 2:
        connection, name = segments
    elif len(segments) == 1:
        connection = "file_uploads"
        name = segments[0]
    else:
        raise ValueError(
            f"Got invalid dataset reference {s} - expected either <dataset_name> or "
            f"<connection_name>/<dataset_name>"
        )

    return connection, name


def camel_case_from_snake_case(s: str) -> str:
    # https://www.geeksforgeeks.org/python-convert-snake-case-string-to-camel-case/
    init, *temp = s.split("_")

    # using map() to get all words other than 1st
    # and titlecasing them
    res = "".join([init.lower(), *map(str.title, temp)])
    return res


def strip_api_from_gateway_url(url: str) -> str:
    """Strips 'api.' from the URL if it starts with 'https://api.'"""
    # Check if the URL starts with "https://api"
    if url.startswith("https://api"):
        # Replace "https://api" with "https://"
        return url.replace("https://api.", "https://", 1)
    return url


def validate_openai_base_model_support(base_model: str, client: openai.Client) -> None:
    """Validates that the user specified base model to use for augmentation is a valid OpenAI model
    and one that we support for augmentation.

    Inputs:
    :param base_model: (str) The OpenAI base model.
    :param client: (openai.Client) The authenticated OpenAI client.

    Returns:
    :return: None
    :raises: ValueError if the model is not supported by OpenAI.
    :raises: ValueError if the model isn't supported by Predibase.
    """
    is_valid_model = any(model_info.id == base_model for model_info in client.models.list().data)
    if not is_valid_model:
        raise ValueError(
            f"The model '{base_model}' is not a valid OpenAI model. Please pass a valid model name.",
        )

    supported_openai_models = {"gpt-4-1106-preview", "gpt-4-0125-preview", "gpt-4-turbo", "gpt-4o", "gpt-4o-mini"}
    if base_model not in supported_openai_models:
        raise ValueError(
            f"The model '{base_model}' is not supported by Predibase. Choose from {supported_openai_models}.",
        )


def validate_openai_api_key(openai_api_key: str) -> openai.Client:
    """Validates that an OpenAI API key is provided and is valid.

    Inputs:
    :param openai_api_key: (str) The OpenAI API key.

    Returns:
    :return: (openai.Client) The authenticated OpenAI client.

    Raises:
    :raises: ValueError if an invalid OpenAI API key is provided.
    :raises: ValueError if an OpenAI API key is not provided.
    """
    # Validate that an OpenAI API key is provided and is valid
    try:
        import openai

        client = openai.Client(api_key=openai_api_key)
        client.models.list()
    except openai.AuthenticationError:
        raise ValueError(
            "An invalid OpenAI API key was provided. Please pass a valid key through `openai_api_key` "
            "or set the OPENAI_API_KEY in your environment",
        )

    return client



from __future__ import annotations

from pydantic import BaseModel, Field, AliasPath


class Adapter(BaseModel):
    repo: str
    tag: int
    checkpoint: int | None = Field(default=None)
    archived: bool
    base_model: str = Field(validation_alias="baseModel")
    description: str | None = Field(default=None)
    artifact_path: str | None = Field(validation_alias="adapterPath", default=None)
    finetuning_error: str | None = Field(validation_alias=AliasPath("finetuningJob", "error"), default=None)
    finetuning_job_uuid: str | None = Field(validation_alias=AliasPath("finetuningJob", "uuid"), default=None)

    @property
    def name(self) -> str:
        return f"{self.repo}/{self.tag}"



from requests.adapters import HTTPAdapter

DEFAULT_TIMEOUT = 5  # seconds


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)



from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING

import requests
from pydantic import BaseModel, Field, field_validator

from predibase._errors import FinetuningError
from predibase.config import FinetuningConfig
from predibase.resources.adapter import Adapter
from predibase.resources.dataset import Dataset
from predibase.resources.repo import Repo

# from predibase.resources.model import PretrainedHuggingFaceLLM, FinetunedLLMAdapter

if TYPE_CHECKING:
    from predibase import Predibase

_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_TEMPLATE_DIR = os.path.join(_PATH_HERE, "../resource/llm/templates")
_CONFIG_FILENAME = "config.yaml"  # Name of config file for loading templates.


class Adapters:
    def __init__(self, client: Predibase):
        self._client = client
        self._session = client._session  # Directly using the session in the short term as we transition to v2.

    def create(
        self,
        *,
        config: FinetuningConfig | dict,
        dataset: str | Dataset,
        continue_from_version: str | None = None,
        repo: str | Repo,
        description: str | None = None,
        show_tensorboard: bool = False,
    ) -> Adapter:

        # Always blocking since `watch` hardcoded to True.
        job = self._client.finetuning.jobs.create(
            config=config,
            dataset=dataset,
            continue_from_version=continue_from_version,
            repo=repo,
            description=description,
            watch=True,
            show_tensorboard=show_tensorboard,
        )

        adapter_version_resp = self._client.http_get(f"/v2/repos/{job.target_repo}/version/{job.target_version_tag}")
        adapter = Adapter.model_validate(adapter_version_resp)

        if not adapter.artifact_path:
            raise FinetuningError(adapter.finetuning_error)

        return adapter

    def get(self, adapter_id: str) -> Adapter:
        adapter_version_resp = self._fetch(adapter_id)
        adapter = Adapter.model_validate(adapter_version_resp)

        if adapter_version_resp["status"] and adapter_version_resp["status"] not in (
            "completed",
            "errored",
            "canceled",
        ):
            # Training is still ongoing, so watch the progress and refetch after.
            print(f"Adapter {adapter_id} is not yet ready.")
            self._client.finetuning.jobs.watch(adapter.finetuning_job_uuid)

            adapter_version_resp = self._fetch(adapter_id)
            adapter = Adapter.model_validate(adapter_version_resp)

        if not adapter.artifact_path:
            raise FinetuningError(adapter.finetuning_error or "Unknown error - adapter not available.")

        return adapter

    def cancel(self, adapter_id: str):
        adapter_version_resp = self._fetch(adapter_id)
        adapter = Adapter.model_validate(adapter_version_resp)

        if not adapter.finetuning_job_uuid:
            raise RuntimeError(f"Adapter {adapter_id} is not associated with a cancelable finetuning job.")

        self._client.finetuning.jobs.cancel(adapter.finetuning_job_uuid)

    def archive(self, adapter_id: str):
        repo, version_tag = self._parse_id(adapter_id)
        self._client.http_put(
            f"/v2/repos/{repo}/version/{version_tag}",
            json={
                "archived": True,
            },
        )

    def unarchive(self, adapter_id: str):
        repo, version_tag = self._parse_id(adapter_id)
        self._client.http_put(
            f"/v2/repos/{repo}/version/{version_tag}",
            json={
                "archived": False,
            },
        )

    def delete(self, adapter_id: str):
        repo, version_tag = self._parse_id(adapter_id)
        self._client.http_delete(f"/v2/repos/{repo}/version/{version_tag}")

    def download(self, adapter_id: str, dest: os.PathLike | None = None):
        repo, version_tag = self._parse_id(adapter_id)

        if dest is None:
            dest = os.path.join(os.getcwd(), f"{version_tag}.zip")

        if os.path.isdir(dest):
            dest = os.path.join(dest, f"{version_tag}.zip")

        print(f"Downloading adapter {adapter_id} as {dest}...")
        with self._client._http.get(
            self._client.api_gateway + f"/v2/repos/{repo}/version/{version_tag}/download",
            headers=self._client.default_headers,
        ) as r:
            try:
                r.raise_for_status()
                with open(dest, "wb") as f:
                    f.write(r.content)
                print("Done!")
            except requests.HTTPError as e:
                try:
                    response_text_obj = json.loads(r.text)
                    if "error" in response_text_obj:
                        if "message" in response_text_obj["error"]:
                            raise RuntimeError(response_text_obj["error"]["message"]) from e
                    raise e
                except json.JSONDecodeError as je:
                    raise ValueError(f"Encountered unexpected problem while decoding error: {r.text}") from je

    def upload(self, local_dir: str, repo: str, base_model: str | None = None) -> Adapter:
        if not os.path.isdir(local_dir):
            raise ValueError(f"{local_dir} does not exist or is not a directory.")

        if not os.path.isfile(os.path.join(local_dir, "adapter_config.json")):
            raise ValueError("Required file adapter_config.json is missing or is not a file.")

        if not os.path.isfile(os.path.join(local_dir, "adapter_model.bin")) and not os.path.isfile(
            os.path.join(local_dir, "adapter_model.safetensors"),
        ):
            raise ValueError("At least one of adapter_model.bin or adapter_model.safetensors is required.")

        begin_resp = self._client.http_post(
            "/v2/adapters/upload",
            json={
                "repo": repo,
            },
        )
        upload_info = _BeginAdapterUploadResponse.model_validate(begin_resp)

        print(f"Uploading adapter to repo {repo}...")
        for model_file in ("adapter_config.json", "adapter_model.bin", "adapter_model.safetensors"):
            # Fine to skip here since we've already validated the necessary set of files above.
            # TODO: consider refactoring for clarity
            if not os.path.exists(os.path.join(local_dir, model_file)):
                continue

            presigned_url = upload_info.presigned_urls[model_file]
            headers = {
                "Content-Type": "application/octet-stream",
                **upload_info.required_headers,
            }

            with open(os.path.join(local_dir, model_file), "rb") as f:
                requests.put(presigned_url, data=f, headers=headers).raise_for_status()

        complete_resp = self._client.http_put(
            "/v2/adapters/upload/",
            json={
                "uploadToken": upload_info.upload_token,
            },
        )
        print("Done!")
        return Adapter.model_validate(complete_resp)

    @staticmethod
    def _parse_id(adapter_id: str):
        segments = adapter_id.split("/", 1)
        if len(segments) != 2:
            raise ValueError("Expected adapter reference of the format <repo>/<version>.")

        return segments

    def _fetch(self, adapter_id: str):
        repo, version_tag = self._parse_id(adapter_id)
        adapter_version_endpoint = f"/v2/repos/{repo}/version/{version_tag}"

        return self._client.http_get(adapter_version_endpoint)


class _BeginAdapterUploadResponse(BaseModel):
    presigned_urls: dict[str, str] = Field(validation_alias="presignedUrls")
    upload_token: str = Field(validation_alias="uploadToken")
    required_headers: dict[str, str] = Field(default_factory=dict, validation_alias="requiredHeaders")

    @field_validator("presigned_urls")
    @classmethod
    def all_presigned_urls_present(cls, v: dict[str, str]) -> dict[str, str]:
        for k in ("adapter_config.json", "adapter_model.bin", "adapter_model.safetensors"):
            if k not in v:
                raise ValueError(f"Missing upload URL for required file {k}")
        return v



import re
from string import capwords
from typing import Optional

import typer
from rich.table import Table

from predibase.cli_commands.utils import get_client, get_console, get_repo

app = typer.Typer(no_args_is_help=True)

KEYS = [
    "name",
    "description",
    "modelName",
    "numShards",
    "quantize",
    "deploymentStatus",
    "promptTemplate",
    "created",
    "updated",
    "scaleDownPeriod",
    "engine",
    "errorText",
]


@app.command(help="List existing Large Langage Model (LLM) deployments")
def llms(
    active_only: bool = typer.Option(False, "--active-only", "-a", help="Only list active LLMs"),
):
    client = get_client()
    llms = client.list_llm_deployments(active_only=active_only, print_as_table=False)
    llms = [{k: llm[k] for k in client._simple_table_keys} for llm in llms]

    camelCaseToTitle = lambda s: capwords(re.sub(r"((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))", r" \1", s).strip())
    table_keys = client._simple_table_keys

    table = Table(*[camelCaseToTitle(k) for k in table_keys])
    get_console().print(f"Found {len(llms)} LLMs:")

    for llm in llms:
        table.add_row(*[str(llm[k]) for k in table_keys])
    get_console().print(table)


@app.command(help="List existing models for a given repository")
def models(repo: Optional[str] = None):
    repo = get_repo(repo)
    table = Table("Version", "Description", "UUID")
    for model in repo.list_models():
        table.add_row(str(model.version), model.description, model.uuid)
    get_console().print(table)


@app.command(help="List existing repositories")
def repos():
    table = Table("Name", "UUID")
    repos = get_client().list_model_repos()
    for repo in repos:
        table.add_row(repo.name, repo.uuid)
    get_console().print(table)


if __name__ == "__main__":
    app()



import typer

from predibase.cli_commands.utils import get_client, get_console

app = typer.Typer(no_args_is_help=True)


@app.command(help="Delete Large Langage Model (LLM) deployments")
def llm(
    deployment_name: str = typer.Option(
        None,
        "--deployment-name",
        "-d",
        prompt="Deployment name",
        prompt_required=True,
        help="Name of the deployment",
    ),
):
    client = get_client()

    get_console().print(f"Deleting LLM with deployment name: {deployment_name}")

    client.LLM(f"pb://deployments/{deployment_name}").delete()
    get_console().print("Delete request sent.")


if __name__ == "__main__":
    app()



from typing import Optional

import typer

from predibase.cli_commands.utils import get_client, get_console, get_dataset, get_engine, get_or_create_repo, load_yaml
from predibase.resource.model import ModelRepo

app = typer.Typer(no_args_is_help=True)


@app.command()
def model(
    config: Optional[str] = None,
    dataset: Optional[str] = None,
    repo: Optional[str] = None,
    engine: Optional[str] = None,
    message: Optional[str] = typer.Option(None, "--message", "-m"),
    watch: bool = False,
):
    config = load_yaml(config)
    engine = get_engine(engine)
    repo: ModelRepo = get_or_create_repo(repo)

    if dataset is None:
        # Assume the dataset is the same as the repo head
        md = repo.head().to_draft()
        md.config = config
    else:
        dataset = get_dataset(dataset)
        md = repo.create_draft(config=config, dataset=dataset)

    if message is not None:
        md.description = message

    result = md.train_async(engine=engine)
    get_console().print("Started model training")
    if watch:
        try:
            result.get()
        except KeyboardInterrupt:
            cancel = typer.confirm("Do you wish to cancel the current model training job?")
            if cancel:
                result.cancel()


@app.command()
def engine(
    name: str = typer.Option(
        None,
        "--name",
        "-n",
        prompt="Name of the engine",
        prompt_required=True,
        help="Name of the engine",
    ),
    template: Optional[str] = typer.Option(
        None,
        "--template",
        "-t",
        prompt="Engine template",
        prompt_required=False,
        help="Template for the engine to be created",
    ),
):
    client = get_client()

    get_console().print(f"Creating engine {name} from template {template}")
    client.create_engine(name, template)


if __name__ == "__main__":
    app()



from __future__ import annotations

import logging
import os
import platform
from http import HTTPStatus
from typing import Any

import requests
import websockets
from urllib3 import Retry
from websockets.sync.client import connect

from predibase._errors import PredibaseError, PredibaseResponseError, PredibaseServerError, warn_outdated_sdk
from predibase.pql import start_session, get_session
from predibase.pql.adapter import TimeoutHTTPAdapter
from predibase.resources.adapters import Adapters

# from predibase.resources.completions import Completions
from predibase.resources.datasets import Datasets
from predibase.resources.deployments import Deployments
from predibase.resources.finetuning_jobs import Finetuning

# from predibase.resources.models import Models
from predibase.resources.repos import Repos
from predibase.util import get_serving_endpoint
from predibase.util.settings import load_settings

# from predibase.util.json import JSONType
from predibase.util.util import remove_suffix
from predibase.version import __version__

_DEFAULT_API_GATEWAY = "https://api.app.predibase.com"

logger = logging.getLogger(__name__)


class Predibase:
    def __init__(
        self,
        *,
        api_token: str | None = None,
    ):
        """Create a new Predibase client instance.

        If not provided, optional params will be set by, in order of precedence:
        * Current environment variables
        * Hard-coded defaults

        :param api_token: The authentication token for the Predibase API.
        """
        # TODO: this is still rather buggy in notebooks.
        # if os.getenv("PREDIBASE_DEBUG") != "":
        #     ipython = None
        #     try:
        #         ipython = get_ipython()  # noqa
        #     except NameError:
        #         # We're not in a notebook; use standard tracebacklimit instead.
        #         # IMPORTANT: setting this to a value <= 1 breaks colab horribly and has no effect in Jupyter.
        #         sys.tracebacklimit = 0
        #
        #     # https://stackoverflow.com/a/61077368
        #     if ipython:
        #
        #         def hide_traceback(
        #             exc_tuple=None, filename=None, tb_offset=None, exception_only=False, running_compiled_code=False
        #         ):
        #            etype, value, tb = sys.exc_info()
        #            return ipython._showtraceback(etype, value, ipython.InteractiveTB.get_exception_only(etype, value))
        #
        #         ipython.showtraceback = hide_traceback

        api_token = api_token or os.environ.get("PREDIBASE_API_TOKEN") or load_settings().get("token")
        api_gateway = os.environ.get("PREDIBASE_GATEWAY") or _DEFAULT_API_GATEWAY
        api_gateway = remove_suffix(api_gateway, "/v1")
        api_gateway = remove_suffix(api_gateway, "/v2")

        if not api_token:
            raise PredibaseError(
                "An api_token must either be explicitly provided or set via the PREDIBASE_API_TOKEN "
                "environment variable.",
            )

        if not api_gateway:
            raise PredibaseError("PREDIBASE_GATEWAY cannot be empty.")

        if not api_gateway.startswith("https://") and not api_gateway.startswith("http://"):
            logger.info(
                f"No HTTP scheme provided for PREDIBASE_GATEWAY (got: {api_gateway}), defaulting to `https://`.",
            )
            api_gateway = f"https://{api_gateway}"

        serving_endpoint = get_serving_endpoint(api_gateway)

        self.api_token = api_token
        self.api_gateway = api_gateway
        self.serving_http_endpoint = serving_endpoint

        self._session = start_session(self.api_gateway, self.api_token, self.serving_http_endpoint)

        self.adapters = Adapters(self)
        # self.completions = Completions(self)
        self.datasets = Datasets(self)
        self.deployments = Deployments(self)
        # self.models = Models(self._session)
        self.repos = Repos(self)
        self.finetuning = Finetuning(self)

    @property
    def default_headers(self) -> dict[str, str]:
        return {
            "Authorization": "Bearer " + self.api_token,
            "User-Agent": f"predibase-sdk/{__version__} ({platform.version()})",
        }

    @property
    def _http(self):
        """Returns a http requests configured with back off and timeout. Will timeout after 10 minutes, and will
        sleep for time based on:

        {backoff factor} * (2 ** ({number of total retries} - 1))
        """
        # see: https://findwork.dev/blog/advanced-usage-python-requests-timeouts-retries-hooks/
        retry_strategy = Retry(
            total=4,
            backoff_factor=2,
            status_forcelist=[104, 429, 500, 502, 503, 504],  # Exclude bad request
            allowed_methods=["HEAD", "GET", "PUT", "PATCH", "POST", "OPTIONS"],
            raise_on_status=False,  # Get the last response (which may contain useful error info) instead of unhelpful
            # retry error (e.g., "too many 500 responses").
        )
        session = self._session if self._session else get_session()
        http = session.requests_session
        adapter = TimeoutHTTPAdapter(
            max_retries=retry_strategy,
            timeout=600,
        )
        http.mount("https://", adapter)
        http.mount("http://", adapter)
        return http

    def http_get(self, endpoint: str, params: dict | None = None, **kwargs) -> dict:
        @warn_outdated_sdk
        def do():
            return self._http.get(self.api_gateway + endpoint, headers=self.default_headers, params=params, **kwargs)

        return _to_json(do())

    def http_post(self, endpoint: str, data: Any | None = None, json: Any | None = None, **kwargs) -> dict:
        @warn_outdated_sdk
        def do():
            return self._http.post(
                self.api_gateway + endpoint,
                data=data,
                json=json,
                headers=self.default_headers,
                **kwargs,
            )

        return _to_json(do())

    def http_delete(self, endpoint: str) -> dict:
        @warn_outdated_sdk
        def do():
            return self._http.delete(self.api_gateway + endpoint, headers=self.default_headers)

        return _to_json(do())

    def http_put(self, endpoint: str, json: Any | None) -> dict:
        @warn_outdated_sdk
        def do():
            return self._http.put(self.api_gateway + endpoint, json=json, headers=self.default_headers)

        return _to_json(do())

    def listen_websocket(self, endpoint: str, max_attempts: int = 10) -> websockets.Data:
        ws_url = self.api_gateway.replace("https://", "wss://").replace("http://", "ws://")

        attempts = 0
        conn = None
        while attempts < max_attempts:
            attempts += 1
            try:
                conn = connect(ws_url + endpoint, additional_headers=self.default_headers)

                while True:
                    try:
                        yield conn.recv()
                    except websockets.exceptions.ConnectionClosedError:
                        break
                    except websockets.exceptions.ConnectionClosedOK:
                        return

            finally:
                if conn is not None:
                    conn.close()

        raise RuntimeError("TODO populate")


def _to_json(resp: requests.Response) -> dict:
    if 200 <= resp.status_code < 400:
        return {} if resp.status_code == 204 else payload_json(resp)

    if 400 <= resp.status_code < 500:
        payload = payload_json(resp)
        raise RuntimeError(
            f"Bad request. Response status code {resp.status_code}. Error: " f"{payload.get('error', 'Unknown')}",
        )

    if 500 <= resp.status_code < 600:
        payload = payload_json(resp)
        raise PredibaseServerError(
            f"Server error. Response status code {resp.status_code}. Error: " f"{payload.get('error', 'Unknown')}",
        )

    raise PredibaseResponseError(f"Unexpected response code: {resp.status_code}", resp.status_code)


def payload_json(r) -> dict:
    try:
        data = r.json()

        if not isinstance(data, dict):
            # New APIs should always return a JSON object.
            raise PredibaseResponseError(
                f"Invalid response payload - expected a JSON object, got {data}",
                HTTPStatus.INTERNAL_SERVER_ERROR,
            )

        return data

    except requests.exceptions.JSONDecodeError as e:
        raise PredibaseResponseError(
            f"Failed to decode payload as JSON. Response status code: "
            f"{r.status_code}. Raw payload text: \n{r.text}. Error: {e}\n",
            r.status_code,
        ) from e



from dataclasses import dataclass, field
from typing import List, Optional

import pandas as pd
from dataclasses_json import config, dataclass_json, LetterCase

from predibase.pql.api import ServerResponseError, Session
from predibase.resource.connection_object import ConnectionObject
from predibase.resource.connection_properties import ConnectionType
from predibase.resource.user import User
from predibase.util import log_info


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Connection:
    session: Session
    _id: str = field(metadata=config(field_name="id"))
    _name: str = field(metadata=config(field_name="name"))
    _connection_type: ConnectionType = field(metadata=config(field_name="type"))
    _user: Optional[User] = field(metadata=config(field_name="user"), default=None)

    def __repr__(self):
        return (
            f"Connection(id={self.id}, name={self.name}, connection_type={self.connection_type}, author={self.author})"
        )

    def _build_dataset_with_backref(self, resp):
        from predibase.resource.dataset import Dataset

        resp["connection"] = self
        return Dataset.from_dict({"session": self.session, **resp})

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def author(self):
        return self._user.username if self._user else None

    @property
    def connection_type(self):
        return self._connection_type

    """List datasets associated with connection"""

    def list_datasets(self, df: bool = False) -> List[dict]:
        resp = self.session.get_json(f"/connections/{self.id}?with_datasets=true")
        datasets = resp["datasets"]
        if df:
            return pd.DataFrame(datasets).drop(columns=["connection"])
        return [self._build_dataset_with_backref(x) for x in datasets]

    """
    List possible objects (files, tables, etc.) that can be imported as datasets
    For flat file credentials (S3, GCS, etc.) this will return an empty list
    """

    def list_objects(self, df: bool = False) -> List[str]:
        resp = self.session.get_json(f"/connections/{self.id}/schema")
        if df:
            return pd.DataFrame(resp["tables"])
        return [ConnectionObject.from_dict(x) for x in resp["tables"]]

    def get_dataset(
        self,
        dataset_name: Optional[str] = None,
        dataset_id: Optional[int] = None,
    ) -> "Dataset":  # noqa
        return get_dataset(self.session, dataset_name, self.name, dataset_id)

    def import_dataset(self, object_name: str, dataset_name: str, exists_ok: bool = False) -> "Dataset":  # noqa
        if not self.session.is_plan_expired():
            req = {
                "connectionID": self.id,
                # "createDatasetsNames": [DatasetNaming(object_name, dataset_name)]
                "createDatasetsNames": [{"objectName": object_name, "datasetName": dataset_name}],
            }

            try:
                resp = self.session.post_json("/datasets", req)
            except ServerResponseError as e:
                if exists_ok and e.code == 400:
                    log_info(f"Dataset {dataset_name} already exists. exists_ok=True, so ignoring.")
                    return self.get_dataset(dataset_name)
                else:
                    raise e

            # TODO(hungcs): don't wait for dataset but move dataset creation outside workflow to be synchronous
            endpoint = f"/datasets/name/{dataset_name}?connectionName={self.name}"
            resp = self.session.wait_for_dataset(endpoint, until_fully_connected=True)
            return self._build_dataset_with_backref(resp)
        else:
            raise PermissionError(
                "Connecting data is locked for expired plans. Contact us to upgrade.",
            )


def get_dataset(
    session: Session,
    dataset_name: Optional[str] = None,
    connection_name: Optional[str] = None,
    dataset_id: Optional[int] = None,
) -> "Dataset":  # noqa
    from predibase.resource_util import build_dataset

    if not (dataset_name or connection_name or dataset_id):
        raise ValueError("Must provide one of [dataset_name, dataset_name + connection_name, dataset_id]")
    if dataset_id:
        if dataset_name or connection_name:
            raise ValueError("Cannot provide both dataset_id and dataset_name/connection_name")
        endpoint = f"/datasets/{dataset_id}"
    else:
        if connection_name and not dataset_name:
            raise ValueError("Cannot provide connection_name without dataset_name")
        endpoint = f"/datasets/name/{dataset_name}"
    if connection_name:
        endpoint += "?connectionName=" + connection_name
    resp = session.get_json(endpoint)

    return build_dataset(resp, session)



import os
import sys
import subprocess
import tempfile
from threading import Event, Thread
from typing import Optional

import typer
from rich import box
from rich.table import Table

# from predibase.cli_commands import run
from predibase.cli_commands import create, delete, deploy, download, finetune, list_resources, prompt, settings, upload
from predibase.cli_commands.utils import df_to_table, get_client, get_console, set_defaults_from_settings
from predibase.util.metrics import model_logs_directory
from predibase.util.settings import load_settings, save_global_settings, save_local_settings

if not os.getenv("PREDIBASE_ENABLE_TRACEBACK"):
    sys.tracebacklimit = 0

app = typer.Typer(help="Predibase CLI commands", no_args_is_help=True)

app.add_typer(create.app, name="create", help="Create Predibase resources")
app.add_typer(deploy.app, name="deploy", help="Deploy Predibase resources")
app.add_typer(delete.app, name="delete", help="Delete Predibase resources")
app.add_typer(list_resources.app, name="list", help="List Predibase resources")
app.add_typer(prompt.app, name="prompt", help="Prompt Predibase models")
app.add_typer(settings.app, name="settings", help="Configure Predibase settings")
app.add_typer(upload.app, name="upload", help="Upload to Predibase")
app.add_typer(finetune.app, name="finetune", help="Finetune a model")
app.add_typer(download.app, name="download", help="Download an artifact")


@app.command(help="Get Predibase SDK version")
def version():
    import predibase

    get_console().print(f"Predibase SDK, version: '{predibase.__version__}'")


@app.command(help="Login to Predibase using API token")
def login(
    token: Optional[str] = typer.Option(
        None,
        "--token",
        "-t",
        help="The optional api token",
        prompt="API Token",
        hide_input=True,
    ),
):
    settings = load_settings()
    settings["token"] = token
    save_global_settings(settings)
    set_defaults_from_settings(settings)

    try:
        user = get_client().get_current_user()
        get_console().print(f"\n🚀 Welcome to Predibase, '{user.username}'!")
    except Exception:
        raise RuntimeError("❌ Unable to login to Predibase! Is your API token correct?")


@app.command(help="Initialize a profile to store default model repository and engine")
def init_profile(
    repository_name: Optional[str] = typer.Option(
        None,
        "--repository-name",
        "-r",
        help="The optional model repository name",
    ),
    engine_name: Optional[str] = typer.Option(None, "--engine-name", "-e", help="The optional engine name"),
    quiet: bool = False,
):
    save_local_settings({k: v for k, v in dict(repo=repository_name or "", engine=engine_name or "").items() if v})


@app.command(help="Train a new model from a Ludwig config")
def train(
    config: Optional[str] = None,
    dataset: Optional[str] = None,
    repo: Optional[str] = None,
    engine: Optional[str] = None,
    message: Optional[str] = typer.Option(None, "--message", "-m"),
    watch: bool = False,
):
    create.model(config, dataset, repo, engine, message, watch)


@app.command(help="Predict using a trained model")
def predict(
    model_name: str = typer.Option(
        None,
        "--model-name",
        "-m",
        prompt="Name of the model to predict with",
        prompt_required=True,
        help="Name of the model to predict with",
    ),
    target: str = typer.Option(
        None,
        "--target",
        prompt="Prediction target",
        prompt_required=True,
        help="Prediction target column",
    ),
    input_csv: Optional[str] = typer.Option(
        None,
        "--input-csv",
        prompt="CSV file containing data to predict on",
        prompt_required=True,
        help="CSV file containing data to predict on",
    ),
    engine_name: Optional[str] = typer.Option(
        None,
        "--engine",
        "-e",
        prompt="Engine name",
        prompt_required=False,
        help="Engine to use for running the predict",
    ),
):
    client = get_client()

    if "/" in model_name:
        repo, version = model_name.split("/")
        model = client.get_model(repo, version)
    else:
        model = client.get_model(model_name)

    engine = client.get_engine(engine_name) if engine_name is not None else None

    import pandas as pd

    df = pd.read_csv(input_csv)

    res = model.predict(targets=target, source=df, engine=engine)

    table = Table(show_header=True, header_style="bold magenta")

    # Modify the table instance to have the data from the DataFrame
    table = df_to_table(res, table)

    # Update the style of the table
    table.row_styles = ["none", "dim"]
    table.box = box.SIMPLE_HEAD

    get_console().print(table)


@app.command(help="Launch a TensorBoard instance")
def tensorboard(
    model_id: int = typer.Option(
        None,
        "--model-id",
        "-m",
        prompt="Model ID",
        prompt_required=True,
        help="The model ID to launch a TensorBoard for",
    ),
):
    client = get_client()

    get_console().print(f"Launching TensorBoard for Model {model_id}")

    # Check if the model has logs available or will have logs available
    model_logs = client.session.get_model_logs_urls(model_id)
    if len(model_logs.get("logfiles", [])) == 0 and not model_logs.get("moreLogsAvailable", False):
        get_console().print(
            f"Model {model_id} does not have logs available, which is likely because the model failed or was terminated before logs were created. Please check model status.",  # noqa: E501
        )
        return

    with tempfile.TemporaryDirectory() as tmpdir:
        logs_dir_name = os.path.join(tmpdir, model_logs_directory(model_id))
        tensorboard_event = Event()
        logs_sync_thread = Thread(
            target=client.session.get_model_logs,
            args=[model_id, tensorboard_event, None, logs_dir_name],
            daemon=True,  # TODO: Figure out why Signal isn't working...
        )
        logs_sync_thread.start()

        subprocess.run(["tensorboard", f"--logdir={logs_dir_name}"])

        # Stop syncing logs when TensorBoard is closed
        tensorboard_event.set()
        logs_sync_thread.join()


def main():
    set_defaults_from_settings(load_settings())
    app()


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from typing import List

from dataclasses_json import config, dataclass_json, LetterCase

# Copied from https://github.com/ludwig-ai/ludwig/blob/master/ludwig/automl/base_config.py


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class FieldInfo:
    name: str
    dtype: str
    key: str = None
    distinct_values: List = None
    distinct_values_balance: float = 1.0
    num_distinct_values: int = 0
    nonnull_values: int = 0
    image_values: int = 0
    audio_values: int = 0
    avg_words: int = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DatasetInfo:
    fields: List[FieldInfo]
    row_count: int
    size_bytes: int = -1


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DatasetInfoWrapper:
    dataset_id: int = field(metadata=config(field_name="datasetID"))
    info: DatasetInfo = field(metadata=config(field_name="info"))



from __future__ import annotations

# import json
import mimetypes
import os
import pathlib
import tempfile
from os import PathLike
from typing import TYPE_CHECKING

import pandas as pd
import requests
from google.protobuf.json_format import MessageToJson
from predibase_api.artifacts.v1.artifacts_pb2 import (
    ARTIFACT_TYPE_DATASET,
    PresignedUrlForUploadRequest,
    PresignedUrlForUploadResponse,
    RegisterUploadedFileAsDatasetRequest,
)

from predibase.config import AugmentationConfig
from predibase.resources.dataset import Dataset
from predibase.resources.util import (
    parse_connection_and_dataset_name,
    strip_api_from_gateway_url,
    validate_openai_api_key,
    validate_openai_base_model_support,
)

# from predibase.resource.dataset import Dataset

if TYPE_CHECKING:
    from predibase import Predibase


class Datasets:
    def __init__(self, client: Predibase):
        self._client = client
        self._session = client._session

    def get(self, dataset_ref: str | Dataset) -> Dataset:
        if isinstance(dataset_ref, Dataset):
            dataset_ref = f"{dataset_ref.connection_name}/{dataset_ref.name}"

        connection, name = parse_connection_and_dataset_name(dataset_ref)

        dataset_resp = self._client.http_get(f"/v1/datasets/name/{name}?connectionName={connection}")
        return Dataset.model_validate(dataset_resp)

    def download(self, dataset_ref: str | Dataset, *, dest: str):
        if isinstance(dataset_ref, Dataset):
            dataset_ref = dataset_ref.uuid

        resp = self._client.http_get(f"/v2/datasets/{dataset_ref}/download-link")

        extension = resp.get("fileExtension", "")
        if extension != "" and not dest.endswith(f".{extension}"):
            dest = dest + f".{extension}"

        r = requests.get(resp["url"])
        r.raise_for_status()

        with open(dest, "wb") as f:
            f.write(r.content)

    def from_file(self, file_path: PathLike, *, name: str | None = None) -> Dataset:
        """Connects the specified file as a Predibase Dataset for training.

        # Inputs
        :param file_path: (str) The file path name for the training dataset (columns "prompt" and "completion" are
            expected).
        :param name: (str) Optional name of the dataset (default is None).
        :return: (Dataset) The connected Predibase Dataset object.
        """
        if name is None:
            name = pathlib.Path(file_path).stem

        with open(file_path, "rb") as f:
            file_name = os.path.basename(file_path)
            mime_type = mimetypes.guess_type(file_path)[0] or "text/plain"

            # Get presigned url for upload
            presigned_url_req = MessageToJson(
                PresignedUrlForUploadRequest(
                    mime_type=mime_type,
                    expiry_seconds=3600,
                    artifact_type=ARTIFACT_TYPE_DATASET,
                ),
                preserving_proto_field_name=True,
            )
            resp = PresignedUrlForUploadResponse(
                **self._session.post(
                    "/datasets/get_presigned_url",
                    data=presigned_url_req,
                ),
            )

            # Unpack response data
            presigned_url = resp.url
            object_name = resp.object_name
            required_headers = resp.required_headers

            # Get required headers for presigned url upload
            headers = {"Content-Type": mime_type}
            for k, v in required_headers:
                headers[k] = v

            # Upload file to blob storage with pre-signed url
            requests.put(presigned_url, data=f, headers=headers).raise_for_status()

            # Register uploaded file as dataset
            register_dataset_req = MessageToJson(
                RegisterUploadedFileAsDatasetRequest(
                    dataset_name=name,
                    object_name=object_name,
                    file_name=file_name,
                ),
                preserving_proto_field_name=True,
            )
            resp = self._session.post("/datasets/register_uploaded_file", data=register_dataset_req)

            dataset_id = resp["id"]
            endpoint = f"/datasets/{dataset_id}?withInfo=true"
            resp = self._session.wait_for_dataset(endpoint, until_fully_connected=True)
            return Dataset.model_validate(resp)

    def from_pandas_dataframe(self, df: pd.DataFrame, *, name: str | None = None) -> Dataset:
        """Connects the specified Pandas DataFrame as a Predibase Dataset for training.

        # Inputs
        :param df: (pd.DataFrame) The Pandas DataFrame reference for the training dataset (columns "prompt" and
            "completion" are expected).
        :param name: (str) Optional name of the dataset (default is None).
        :return: (Dataset) The connected Predibase Dataset object.
        """
        # The file "temp_file" will be automatically deleted when context exits.
        with tempfile.NamedTemporaryFile(suffix=".csv") as temp_file:
            # Write Pandas DataFrame to the temporary file.
            df.to_csv(path_or_buf=temp_file.name, index=False)
            return self.from_file(file_path=temp_file.name, name=name)

    def augment(
        self,
        config: dict | AugmentationConfig,
        dataset: Dataset,
        name: str | None = None,
        openai_api_key: str | None = None,
    ) -> Dataset:
        """Augments the specified DataFrame using the specified configuration.

        # Inputs
        :param config: (dict) The configuration for the augmentation.
        :param dataset: (Dataset) The dataset to augment.
        :param name: (str) The name of the generated dataset, must be unique in the connection.
        :param openai_api_key: (str) The OpenAI API key. If not provided, will be read from `OPENAI_API_KEY`.

        # Returns
        :return: (Dataset) The augmented dataset object.
        """
        if isinstance(config, dict):
            config = AugmentationConfig(**config)

        if openai_api_key is None:
            openai_api_key = os.environ.get("OPENAI_API_KEY")
            if openai_api_key is None:
                raise ValueError(
                    "An OpenAI API key must be provided to augment a dataset. Please pass the key as an "
                    "argument through `openai_api_key` or set the OPENAI_API_KEY environment variable.",
                )

        client = validate_openai_api_key(openai_api_key)
        validate_openai_base_model_support(config.base_model, client)

        augment_response = self._client.http_post(
            f"/v2/datasets/{dataset.uuid}/augment",
            json={
                "name": name,
                "config": dict(config),
                "openaiApiKey": openai_api_key,
            },
        )

        augmented_dataset_id = augment_response["datasetId"]
        base_url = strip_api_from_gateway_url(self._session.url)
        augment_dataset_url = f"{base_url}/data/datasets/{augmented_dataset_id}"
        print(f"Augmentation started! Check for augmentation status at {augment_dataset_url}")

        url = f"/datasets/{augmented_dataset_id}?withInfo=true"
        augmented_dataset = self._session.wait_for_dataset(url, until_fully_connected=True)
        return Dataset.model_validate(augmented_dataset)



import os
import sys
from typing import Optional, Union

from predibase.connection_mixin import ConnectionMixin
from predibase.dataset_mixin import DatasetMixin
from predibase.deployment_mixin import DeploymentMixin
from predibase.engine_mixin import EngineMixin
from predibase.llm_mixin import LlmMixin
from predibase.model_mixin import ModelMixin
from predibase.permission_mixin import PermissionMixin
from predibase.pql import start_session
from predibase.pql.api import Session
from predibase.query_mixin import QueryMixin


class PredibaseClient(
    QueryMixin,
    ConnectionMixin,
    DatasetMixin,
    EngineMixin,
    ModelMixin,
    DeploymentMixin,
    PermissionMixin,
    LlmMixin,
):
    def __init__(self, session: Optional[Session] = None, gateway: str = None, token: str = None):
        self._session = session or start_session(gateway, token)
        if not os.getenv("PREDIBASE_ENABLE_TRACEBACK"):
            ipython = None
            try:
                ipython = get_ipython()  # noqa
            except NameError:
                # We're not in a notebook; use standard tracebacklimit instead.
                # IMPORTANT: setting this to a value <= 1 breaks colab horribly and has no effect in Jupyter.
                sys.tracebacklimit = 0

            # https://stackoverflow.com/a/61077368
            if ipython:

                def hide_traceback(
                    exc_tuple=None, filename=None, tb_offset=None, exception_only=False, running_compiled_code=False
                ):
                    etype, value, tb = sys.exc_info()
                    return ipython._showtraceback(etype, value, ipython.InteractiveTB.get_exception_only(etype, value))

                ipython.showtraceback = hide_traceback

    @property
    def session(self) -> Session:
        return self._session

    def set_connection(self, connection: Union[str, int]):
        """This method sets the default connection to use for any method requiring a connection argument.

        :param connection: The desired connection name or id
        """
        self.session.set_connection(connection)



import re


class PromptTemplate:
    """
    Utility class to make working with prompt templates easier, particularly in the presence of JSON.
    """

    def __init__(self, tmpl: str):
        self._raw = tmpl

    def render(self) -> str:
        escaped = self._raw.replace("{", "{{").replace("}", "}}")

        # Replace $feature with {feature}
        return re.sub(r"\$([A-Za-z0-9_-]+)", r"{\1}", escaped)



import os
from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

extra_requirements = {"notebook": ["notebook", "jupyterlab"]}

with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    requirements = [line.strip() for line in f if line]

with open(path.join(here, "requirements_predictor.txt"), encoding="utf-8") as f:
    extra_requirements["predictor"] = [line.strip() for line in f if line]

version = {}
with open(os.path.join(here, "predibase", "version.py")) as fp:
    exec(fp.read(), version)

setup(
    name="predibase",
    version=version["__version__"],
    author="Predibase Inc.",
    packages=find_packages(),
    zip_safe=False,
    install_requires=requirements,
    extras_require=extra_requirements,
    entry_points={"console_scripts": ["pbase=predibase.cli:main"]},
    include_package_data=True,
)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field, field_validator


class FinetuningConfig(BaseModel):
    # This is required in most cases, but coded as optional for continued training.
    base_model: str | None = Field(default=None)

    adapter: str | None = Field(default=None)
    task: str | None = Field(default=None)
    epochs: int | None = Field(default=None)
    learning_rate: float | None = Field(default=None)
    rank: int | None = Field(default=None)
    target_modules: list[str] | None = Field(default=None)
    enable_early_stopping: bool | None = Field(default=None)
    apply_chat_template: bool | None = Field(default=None)


class UpdateDeploymentConfig(BaseModel):
    # Note: necessary because this particular model and its children are used to both deserialize and serialize data.
    # Other models in this SDK are used exclusively for deserializing data to a Python object that is never (for now)
    # reserialized back to the API for a different request.
    model_config = ConfigDict(populate_by_name=True)

    custom_args: list[str] | None = Field(
        default=None,
        validation_alias="customArgs",
        serialization_alias="customArgs",
    )
    cooldown_time: int | None = Field(
        default=None,
        validation_alias="cooldownTime",
        serialization_alias="cooldownTime",
    )
    hf_token: str | None = Field(
        default=None,
        validation_alias="hfToken",
        serialization_alias="hfToken",
    )
    min_replicas: int | None = Field(
        default=None,
        validation_alias="minReplicas",
        serialization_alias="minReplicas",
    )
    max_replicas: int | None = Field(
        default=None,
        validation_alias="maxReplicas",
        serialization_alias="maxReplicas",
    )
    scale_up_threshold: int | None = Field(
        default=None,
        validation_alias="scaleUpRequestThreshold",
        serialization_alias="scaleUpRequestThreshold",
    )
    lorax_image_tag: str | None = Field(
        default=None,
        validation_alias="loraxImageTag",
        serialization_alias="loraxImageTag",
    )
    request_logging_enabled: bool | None = Field(
        default=None,
        validation_alias="requestLoggingEnabled",
        serialization_alias="requestLoggingEnabled",
    )


class DeploymentConfig(UpdateDeploymentConfig):
    base_model: str = Field(
        ...,
        validation_alias="baseModel",
        serialization_alias="baseModel",
    )
    accelerator: str | None = Field(default=None)
    uses_guaranteed_capacity: bool | None = Field(
        default=None,
        validation_alias="usesGuaranteedCapacity",
        serialization_alias="usesGuaranteedCapacity",
    )
    quantization: str | None = Field(default=None)


class AugmentationConfig(BaseModel):
    """Configuration for synthetic data generation tasks.

    # Attributes
    :param base_model: (str) The OpenAI model to prompt.
    :param num_samples_to_generate: (int) The number of synthetic examples to generate.
    :param num_seed_samples: (int) The number of seed samples to use for generating synthetic examples.
    :param task_context: (str) The user-provided task context for generating candidates.
    """

    base_model: str
    num_samples_to_generate: int = Field(default=1000)
    num_seed_samples: int | str = Field(default="all")
    augmentation_strategy: str = Field(default="mixture_of_agents")
    task_context: str = Field(default="")

    @field_validator("base_model")
    @classmethod
    def validate_base_model(cls, base_model) -> str:
        supoorted_base_models = {
            "gpt-4-turbo",
            "gpt-4-0125-preview",
            "gpt-4-1106-preview",
            "gpt-4o",
            "gpt-4o-2024-08-06",
            "gpt-4o-mini",
        }
        if base_model not in supoorted_base_models:
            raise ValueError(
                f"base_model must be one of {supoorted_base_models}.",
            )
        return base_model

    @field_validator("num_samples_to_generate")
    @classmethod
    def validate_num_samples_to_generate(cls, num_samples_to_generate) -> int:
        if num_samples_to_generate < 1:
            raise ValueError("num_samples_to_generate must be >= 1.")
        return num_samples_to_generate

    @field_validator("num_seed_samples")
    @classmethod
    def validate_num_seed_samples(cls, num_seed_samples):
        if isinstance(num_seed_samples, str) and num_seed_samples != "all":
            raise ValueError("num_seed_samples can only be an integer or the string 'all'.")
        elif isinstance(num_seed_samples, int) and num_seed_samples < 1:
            raise ValueError("num_seed_samples must be >= 1.")
        return num_seed_samples

    @field_validator("augmentation_strategy")
    @classmethod
    def validate_augmentation_strategy(cls, augmentation_strategy):
        if augmentation_strategy not in {"single_pass", "mixture_of_agents"}:
            raise ValueError("augmentation_strategy must be 'single_pass' or 'mixture_of_agents'.")
        return augmentation_strategy



from __future__ import annotations

import datetime
import json
import os
import pathlib
import sys
import tempfile
import time
import urllib.request
from threading import Event, Thread
from typing import Any, Dict, TYPE_CHECKING

from dateutil import parser
from progress_table import ProgressTable
from tqdm import tqdm

from predibase._errors import PredibaseServerError, PredibaseResponseError
from predibase.config import FinetuningConfig
from predibase.resources.dataset import Dataset
from predibase.resources.finetuning_job import FinetuningJob, _FinetuningMetrics
from predibase.resources.repo import Repo

if TYPE_CHECKING:
    from predibase import Predibase


class FinetuningJobs:
    def __init__(self, client: Predibase):
        self._client = client

    def create(
        self,
        *,
        config: FinetuningConfig | Dict[str, Any],
        dataset: str | Dataset,
        continue_from_version: str | None = None,
        repo: str | Repo,
        description: str | None = None,
        watch: bool = False,
        show_tensorboard: bool = False,
    ) -> FinetuningJob:

        if show_tensorboard and not watch:
            raise RuntimeError("`show_tensorboard` is a blocking option and thus requires `watch` to be True.")

        if isinstance(config, FinetuningConfig):
            config = config.dict(exclude_none=True)

        if isinstance(dataset, Dataset):
            dataset = dataset.name  # TODO: eventually accept connection type

        if isinstance(repo, Repo):
            repo = repo.name

        job_resp = self._client.http_post(
            "/v2/finetuning/jobs",
            json={
                "params": config,
                "dataset": dataset,
                "continueFromVersion": continue_from_version,
                "repo": repo,
                "description": description,
            },
        )
        job = FinetuningJob.model_validate(job_resp)

        print(
            f"Successfully requested finetuning of {job.base_model} from "
            f"{'base' if continue_from_version is None else continue_from_version} as `{job.target_repo}/"
            f"{job.target_version_tag}`. (Job UUID: {job.uuid}).\n"
        )

        return self.watch(job, show_tensorboard) if watch else job

    def watch(self, job_ref: str | FinetuningJob, show_tensorboard: bool = False) -> FinetuningJob:
        if isinstance(job_ref, FinetuningJob):
            job_ref = job_ref.uuid

        print(
            f"Watching progress of finetuning job {job_ref}. This call will block until the job has finished. "
            f"Canceling or terminating this call will NOT cancel or terminate the job itself.\n"
        )

        queued_pbar = None
        tensorboard_event = Event()

        # Wait for job to move to the training phase.
        while True:
            try:
                job_resp = self._client.http_get(f"/v2/finetuning/jobs/{job_ref}")
                job = FinetuningJob.model_validate(job_resp)

                # TODO: clean this up
                timeline_resp = self._client.http_get(f"/v2/finetuning/jobs/{job_ref}/timeline")

                # TODO: should an exception be raised in case of failure? Probably not?
                if job.status in ("completed", "canceled", "errored", "stopped", "stopping"):
                    # TODO: print last metrics?
                    print(f"Job {job_ref} is already {job.status}. Nothing to watch.")
                    return job

                if job.status == "queued":
                    msg = f"Job is queued for execution. Time in queue: {_get_elapsed(job.status, timeline_resp)}"
                    if queued_pbar is None:
                        queued_pbar = tqdm(
                            None,
                            bar_format=msg,
                            desc=job.status.capitalize(),
                            ncols=0,
                            file=sys.stdout,
                        )
                    elif not queued_pbar.disable:
                        queued_pbar.bar_format = msg
                        queued_pbar.refresh()
                        queued_pbar.update()

                if job.status == "training":
                    if queued_pbar is not None:
                        queued_pbar.bar_format = (
                            f"Job is starting. Total queue time: {_get_elapsed('queued', timeline_resp)}"
                        )
                        queued_pbar.close()
                    break

            except (PredibaseServerError, PredibaseResponseError) as e:
                if os.getenv("PREDIBASE_DEBUG") != "":
                    print(f"Continuing past error {e}")
                continue

            time.sleep(1)

        # Optionally launch tensorboard and a thread to sync TB logs.
        if show_tensorboard:
            self._run_tensorboard(job_ref, tensorboard_event)

        # Stream metrics during training phase.
        print("Waiting to receive training metrics...\n")
        self._stream_model_metrics(job_ref, _get_metrics_table(), 10)

        # Fetch the final job state after streaming ends.
        while True:
            job_resp = self._client.http_get(f"/v2/finetuning/jobs/{job_ref}")
            job = FinetuningJob.model_validate(job_resp)

            if job.status in ("completed", "canceled", "errored", "stopped", "stopping"):
                if tensorboard_event:
                    tensorboard_event.set()

                return job

            time.sleep(1)

    def _run_tensorboard(self, job_uuid: str, stop_event: Event):
        import tensorboard.notebook

        # tb_logs_resp = self._client.http_get(f"/v2/finetuning/jobs/{job_uuid}/tensorboard_logs")
        # if not tb_logs_resp.get("files", []) and not tb_logs_resp.get("more", False):
        #     print(
        #         f"Finetuning job {job_uuid} does not have logs available, which is likely because the model failed or "
        #         f"was terminated before logs were created.",
        #     )
        #     return

        with tempfile.TemporaryDirectory() as tmpdir:

            def _sync_tb_logfiles():
                while True:
                    tb_logs_resp = self._client.http_get(f"/v2/finetuning/jobs/{job_uuid}/tensorboard_logs")
                    files = tb_logs_resp.get("files", [])
                    should_continue = tb_logs_resp.get("more", False)

                    for f in files:
                        dst = os.path.join(tmpdir, job_uuid, f["filepath"])
                        pathlib.Path(os.path.dirname(dst)).mkdir(parents=True, exist_ok=True)
                        try:
                            urllib.request.urlretrieve(f["url"], dst)
                        except Exception:
                            # TODO: is there a way to safely print errors without breaking the progress table?
                            pass

                    if not should_continue or stop_event.is_set():
                        break

                    time.sleep(30)

            sync_thread = Thread(target=_sync_tb_logfiles, daemon=True)
            sync_thread.start()

            tensorboard.notebook.start(f"--logdir={os.path.join(tmpdir, job_uuid)}")

    def _stream_model_metrics(
        self,
        job_uuid: str,
        table: ProgressTable,
        max_attempts: int = 10,
    ):
        def print_progress_bar(metrics: _FinetuningMetrics):
            if metrics.meta.steps and metrics.meta.steps > 0:
                table.progress_bar_active = True
                table._print_progress_bar(
                    metrics.meta.steps,
                    metrics.meta.total_steps,
                    show_before=f" {metrics.meta.steps}/{metrics.meta.total_steps} steps ",
                )

        # Used to avoid re-printing rows if we need to reconnect, since the server sends historical data for each new
        # connection.
        last_seen_checkpoint = 0

        for resp in self._client.listen_websocket(
            f"/v2/finetuning/jobs/{job_uuid}/metrics/stream", max_attempts=max_attempts
        ):
            try:
                if not table.header_printed:
                    table._print_header(top=True)

                metrics = _FinetuningMetrics.model_validate(json.loads(resp))

                if (
                    metrics.data.steps
                    and metrics.data.steps > 0
                    and metrics.data.checkpoint_number > last_seen_checkpoint
                ):
                    # We've hit an evaluation step / checkpoint. Print a new row entry.
                    table["checkpoint"] = metrics.data.checkpoint_number
                    table["train_loss"] = metrics.data.train_metrics_loss
                    table["validation_loss"] = (
                        metrics.data.validation_metrics_loss if (metrics.data.validation_metrics_loss) else "--"
                    )
                    table.next_row()

                    # Update last seen checkpoint
                    last_seen_checkpoint = metrics.data.checkpoint_number

                print_progress_bar(metrics)

                if metrics.meta.is_completed:
                    table.close()
                    return

            # TODO: Handle specific error types
            except Exception:
                continue

    def get(self, job: str | FinetuningJob) -> FinetuningJob:
        if isinstance(job, FinetuningJob):
            job = job.uuid

        job = self._client.http_get(f"/v2/finetuning/jobs/{job}")
        return FinetuningJob.model_validate(job)

    def cancel(self, job: str | dict):
        if isinstance(job, dict):
            job = job["uuid"]

        self._client.http_post(f"/v2/finetuning/jobs/{job}/cancel")


class Finetuning:
    def __init__(self, client: Predibase):
        self._client = client

        self.jobs = FinetuningJobs(self._client)


def _get_elapsed(curr_status: str, timeline_resp: Dict[str, Any]):
    try:
        started = timeline_resp.get(curr_status, {}).get("startedAt", None)
        start_time = parser.parse(started)

        elapsed = datetime.datetime.utcnow() - start_time.replace(tzinfo=None)
        elapsed = elapsed - datetime.timedelta(microseconds=elapsed.microseconds)

        return str(elapsed)
    except Exception:
        return "{elapsed}"


def _get_metrics_table():
    table = ProgressTable(
        columns=["checkpoint", "train_loss", "validation_loss"],
        num_decimal_places=4,
        reprint_header_every_n_rows=0,
    )
    # Add feature/metric columns separately so we can customize the width
    # table.add_column("feature", width=16)
    # table.add_column("metric", width=24)
    # table.add_columns(["train", "val", "test"])

    # Horrible monkeypatched hack to make progress bar work in colab.
    # TODO: rework this and the logs / metrics streaming code overall.
    def newprint(*args, **kwargs):
        if "end" in kwargs and kwargs["end"] == "\r":
            kwargs["end"] = ""
            args = ("\r",) + args

        for file in table.files:
            print(*args, **kwargs, file=file)

    table._print = newprint

    return table



from __future__ import annotations

import concurrent.futures
import json
import os
import queue
import types
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import requests
from lorax import AsyncClient as AsyncLoraxClient
from lorax import Client as LoraxClient

from predibase.config import DeploymentConfig, UpdateDeploymentConfig
from predibase.resources.deployment import Deployment
from predibase.resources.util import camel_case_from_snake_case

if TYPE_CHECKING:
    from predibase import Predibase

# Mapping from key names in the old dictionary format to the corresponding key in DeploymentConfig format (None if we
# want to drop the field)
dict_to_deployment_config_keys = {
    ".quantize": "quantization",
    ".custom_args": "custom_args",
    ".source": None,
    ".num_shards": None,
    "scale_up_threshold": "scale_up_request_threshold",
}


class Deployments:
    def __init__(self, client: Predibase):
        self._client = client
        self._session = client._session

    # TODO: nameoruuid for get deployment endpoint
    def get(self, deployment_ref: str) -> Deployment:
        dep = self._client.http_get(f"/v2/deployments/{deployment_ref}")["data"]
        return Deployment.model_validate(dep)

    def client(
        self,
        deployment_ref: str | Deployment,
        force_bare_client: bool = False,
        _timeout_in_seconds: int = 600,
        _max_session_retries: int = 2,
    ) -> LoraxClient:
        if isinstance(deployment_ref, Deployment):
            deployment_ref = deployment_ref.name

        if "/" in deployment_ref:
            raise ValueError(
                f"Deployment name {deployment_ref} appears to be invalid. Are you providing a Hugging "
                f"Face path by accident? See https://docs.predibase.com/user-guide/inference/models for "
                f"a list of available deployments.",
            )

        url = (
            f"https://{self._session.serving_http_endpoint}/{self._session.tenant}/deployments/v2/llms/"
            f"{deployment_ref}"
        )

        c = LoraxClient(
            base_url=url,
            headers=self._session.get_headers(),
            timeout=_timeout_in_seconds,
            max_session_retries=_max_session_retries,
        )

        if force_bare_client:
            return c

        c._ready = False

        c._generate = c.generate
        c.generate = types.MethodType(_make_generate(self._client, deployment_ref), c)

        c._generate_stream = c.generate_stream
        c.generate_stream = types.MethodType(_make_generate_stream(self._client, deployment_ref), c)

        return c

    def async_client(self, deployment_ref: str | Deployment, _timeout_in_seconds: int = 600) -> AsyncLoraxClient:
        if isinstance(deployment_ref, Deployment):
            deployment_ref = deployment_ref.name

        url = (
            f"https://{self._session.serving_http_endpoint}/{self._session.tenant}/deployments/v2/llms/"
            f"{deployment_ref}"
        )

        return AsyncLoraxClient(
            base_url=url,
            headers=self._session.get_headers(),
            timeout=_timeout_in_seconds,
        )

    def openai_compatible_endpoint(self, deployment_ref: str | Deployment) -> str:
        if isinstance(deployment_ref, Deployment):
            deployment_ref = deployment_ref.name

        return (
            f"https://{self._session.serving_http_endpoint}/{self._session.tenant}/deployments/v2/llms/"
            f"{deployment_ref}/v1"
        )

    def create(
        self,
        name: str,
        *,
        config: dict | DeploymentConfig,
        description: str | None = None,
    ) -> Deployment:

        # If we have the config in the form of a dictionary, then we should
        # transform the keys into camelCase from snakeCase
        if isinstance(config, dict):
            config = self.convert_dict_to_deployment_config(config)

        if isinstance(config, DeploymentConfig):
            config = config.model_dump(mode="json", by_alias=True)

        payload = {
            "name": name,
            "config": config,
            "description": description or "",
        }

        self._client.http_post(
            "/v2/deployments",
            json=payload,
        )

        self._session.get_llm_deployment_events_until_with_logging(
            events_endpoint=f"/llms/{name}/events?detailed=false",
            success_cond=lambda resp: "Ready" in [r.get("eventType", None) for r in resp.get("ComputeEvents", [])],
            error_cond=lambda resp: "Failed" in [r.get("eventType", None) for r in resp.get("ComputeEvents", [])]
            or resp.get("deploymentStatus", None) in ("failed", "deleted", "stopped"),
        )

        return self.get(name)

    @staticmethod
    def convert_dict_to_deployment_config(config):
        new_config = {}
        for key, value in config.items():
            # Some key names have changed in DeploymentConfig, or we want to ignore entirely.
            key = dict_to_deployment_config_keys.get(key, key)
            if key is None:
                continue
            # Expected to be a map in dict format but a list in DeploymentConfig format
            if key == "custom_args":
                new_value = []
                for k, v in value.items():
                    new_value.append(k)
                    if v != "":
                        new_value.append(v)
                value = new_value

            new_config[camel_case_from_snake_case(key)] = value
        print(f"nc: {new_config}")
        return new_config

    def list(self, *, type: str | None = None) -> list[Deployment]:
        endpoint = "/v2/deployments"

        if type is not None:
            type = type.lower()
            if type not in ("serverless", "dedicated", "shared", "private"):
                raise ValueError("Type filter must be one of `shared` or `private`")

            endpoint = f"{endpoint}?type={type}"

        resp = self._client.http_get(endpoint)
        return [Deployment.model_validate(d) for d in resp["data"]]

    def update(
        self,
        deployment_ref: str | Deployment,
        *,
        description: str | None = None,
        config: UpdateDeploymentConfig | None = None,
    ) -> Deployment:
        if isinstance(deployment_ref, Deployment):
            deployment_ref = deployment_ref.name

        payload = {}
        if config is not None:
            payload["config"] = config.model_dump(mode="json", by_alias=True)

        if description is not None:
            payload["description"] = description

        self._client.http_put(
            f"/v2/deployments/{deployment_ref}",
            json=payload,
        )

        self._session.get_llm_deployment_events_until_with_logging(
            events_endpoint=f"/llms/{deployment_ref}/events?detailed=false",
            success_cond=lambda resp: "Ready" in [r.get("eventType", None) for r in resp.get("ComputeEvents", [])],
            error_cond=lambda resp: "Failed" in [r.get("eventType", None) for r in resp.get("ComputeEvents", [])]
            or resp.get("deploymentStatus", None) in ("failed", "deleted", "stopped"),
        )

    def delete(self, deployment_ref: str | Deployment):
        if isinstance(deployment_ref, Deployment):
            deployment_ref = deployment_ref.name

        self._client.http_delete(f"/v2/deployments/{deployment_ref}")

    def get_recommended_config(
        self,
        base_model: str,
        accelerator: str | None = None,
        quantization: str | None = None,
    ) -> DeploymentConfig:
        if base_model is None:
            raise ValueError("Base model must be specified")

        params = {
            "baseModel": base_model,
        }
        if accelerator is not None:
            params["accelerator"] = accelerator
        if quantization is not None:
            params["quantization"] = quantization

        return self._client.http_get("/v2/deployments/recommended-config", params=params)

    def download_request_logs(
        self,
        deployment_ref: str,
        dest: os.PathLike | None = None,
        *,
        adapter_id: str | None = None,
        from_: str | None = None,
        to: str | None = None,
    ):
        if dest is None:
            dest = os.path.join(os.getcwd(), f"{deployment_ref}_logs.zip")

        if os.path.isdir(dest):
            dest = os.path.join(dest, f"{deployment_ref}_logs.zip")

        if from_ is None:
            from_ = datetime.utcnow().replace(minute=0, second=0, microsecond=0).isoformat()

        if to is None:
            to = datetime.utcnow() + timedelta(hours=1)
            to = to.replace(minute=0, second=0, microsecond=0).isoformat()

        print(f"Downloading logs for deployment {deployment_ref} - {adapter_id} as {dest}...")

        endpoint = self._client.api_gateway + f"/v2/deployments/{deployment_ref}/request-logs?from={from_}&to={to}"
        if adapter_id is not None:
            endpoint += f"&adapterId={adapter_id}"

        with self._client._http.get(endpoint, headers=self._client.default_headers) as r:
            try:
                r.raise_for_status()
                with open(dest, "wb") as f:
                    f.write(r.content)
                print("Done!")
            except requests.HTTPError as e:
                try:
                    response_text_obj = json.loads(r.text)
                    if "error" in response_text_obj:
                        if "message" in response_text_obj["error"]:
                            raise RuntimeError(response_text_obj["error"]["message"]) from e
                    raise e
                except json.JSONDecodeError as je:
                    raise ValueError(f"Encountered unexpected problem while decoding error: {r.text}") from je


def _make_generate(pb: Predibase, deployment_ref: str):
    def _lorax_generate(self, *args, **kwargs):
        if self._ready:
            return self._generate(*args, **kwargs)

        def _generate_thread(q: queue.Queue):
            try:
                q.put_nowait({"type": "generate", "data": self._generate(*args, **kwargs)})
            except Exception as e:
                q.put_nowait({"type": "generate", "exception": e})

        def _status_thread(q: queue.Queue):
            try:
                q.put_nowait({"type": "status", "status": pb.deployments.get(deployment_ref).status})
            except Exception:
                # q.put_nowait({"type": "status", "exception": e})
                pass

        q = queue.Queue()
        with concurrent.futures.ThreadPoolExecutor(max_workers=2) as pool:
            pool.submit(_generate_thread, q)
            pool.submit(_status_thread, q)

            while True:
                resp = q.get()
                if resp["type"] == "generate":
                    if "exception" in resp:
                        raise resp["exception"]

                    self._ready = True
                    return resp["data"]

                if resp["type"] == "status":
                    if resp["status"] not in ("ready", "updating"):
                        print(
                            f"Deployment {deployment_ref} is still spinning up. Your prompt may take longer than "
                            f"normal to execute.\n",
                        )
                    else:
                        self._ready = True

    return _lorax_generate


def _make_generate_stream(pb: Predibase, deployment_ref: str):
    def _lorax_generate_stream(self, *args, **kwargs):
        if self._ready:
            return self._generate_stream(*args, **kwargs)

        def _generate_thread(q: queue.Queue):
            try:
                for r in self._generate_stream(*args, **kwargs):
                    q.put_nowait({"type": "generate", "data": r})

                q.put_nowait(None)
            except Exception as e:
                q.put_nowait({"type": "generate", "exception": e})

        def _status_thread(q: queue.Queue):
            try:
                q.put_nowait({"type": "status", "status": pb.deployments.get(deployment_ref).status})
            except Exception as e:
                q.put_nowait({"type": "status", "exception": e})

        q = queue.Queue()
        resp_seen = False
        with concurrent.futures.ThreadPoolExecutor(max_workers=2) as pool:
            pool.submit(_generate_thread, q)
            pool.submit(_status_thread, q)

            while True:
                resp = q.get()

                if resp is None:
                    break

                if resp["type"] == "generate":
                    if "exception" in resp:
                        raise resp["exception"]

                    resp_seen = True
                    self._ready = True
                    yield resp["data"]
                    continue

                if resp["type"] == "status" and not resp_seen:
                    if resp["status"] not in ("ready", "updating"):
                        print(
                            f"Deployment {deployment_ref} is still spinning up. Your prompt may take longer than "
                            f"normal to execute.\n",
                        )
                    else:
                        self._ready = True

    return _lorax_generate_stream
